#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Constants for the numina package."""


# Factor to obtain FWHM in Gaussian profile from sigma
FWHM_G = 2.3548200450309493


# Square root of two pi
M_SQRT_2_PI = 2.5066282746310005024157652848110452530069867406099383166299
